#include <boost/filesystem.hpp>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <set>
#include <string>

namespace fs = boost::filesystem;

// header -> module
std::map< std::string, std::string > s_header_map;

static void scan_module_headers( fs::path const & path )
{
	try
	{
		std::string module = path.generic_string().substr( 5 ); // "libs/"

		fs::path dir = path / "include";
		size_t n = dir.generic_string().size();

		fs::recursive_directory_iterator it( dir ), last;

		for( ; it != last; ++it )
		{
			fs::directory_entry const & e = *it;

			std::string p2 = e.path().generic_string();
			p2 = p2.substr( n+1 );

			// std::cout << module << ": " << p2 << std::endl;

			s_header_map[ p2 ] = module;
		}
	}
	catch( fs::filesystem_error const & x )
	{
		std::cout << x.what() << std::endl;
	}
}

static void scan_submodules( fs::path const & path )
{
	fs::directory_iterator it( path ), last;

	for( ; it != last; ++it )
	{
		fs::directory_entry const & e = *it;

		if( e.status().type() != fs::directory_file )
		{
			continue;
		}

		fs::path path = e.path();

		if( fs::exists( path / "include" ) )
		{
			scan_module_headers( path );
		}

		if( fs::exists( path / "sublibs" ) )
		{
			scan_submodules( path );
		}
	}
}

static void build_header_map()
{
	scan_submodules( "libs" );
}

static void scan_header_dependencies( std::string const & header, std::istream & is, std::map< std::string, std::set< std::string > > & deps, std::map< std::string, std::set< std::string > > & from )
{
	std::string line;

	while( std::getline( is, line ) )
	{
		while( !line.empty() && ( line[0] == ' ' || line[0] == '\t' ) )
		{
			line.erase( 0, 1 );
		}

		if( line.empty() || line[0] != '#' ) continue;

		line.erase( 0, 1 );

		while( !line.empty() && ( line[0] == ' ' || line[0] == '\t' ) )
		{
			line.erase( 0, 1 );
		}

		if( line.substr( 0, 7 ) != "include" ) continue;

		line.erase( 0, 7 );

		while( !line.empty() && ( line[0] == ' ' || line[0] == '\t' ) )
		{
			line.erase( 0, 1 );
		}

		if( line.size() < 2 ) continue;

		char ch = line[0];

		if( ch != '<' && ch != '"' ) continue;

		if( ch == '<' )
		{
			ch = '>';
		}

		line.erase( 0, 1 );

		std::string::size_type k = line.find_first_of( ch );

		if( k != std::string::npos )
		{
			line.erase( k );
		}

		std::map< std::string, std::string >::const_iterator i = s_header_map.find( line );

		if( i != s_header_map.end() )
		{
			deps[ i->second ].insert( line );
			from[ line ].insert( header );
		}
		else if( line.substr( 0, 6 ) == "boost/" )
		{
			deps[ "*unknown*" ].insert( line );
			from[ line ].insert( header );
		}
	}
}

static void scan_module_dependencies( std::string const & module )
{
	// module -> [ header, header... ]
	std::map< std::string, std::set< std::string > > deps;

	// header -> included from [ header, header... ]
	std::map< std::string, std::set< std::string > > from;

	fs::path dir = fs::path( "libs" ) / module / "include";
	size_t n = dir.generic_string().size();

	fs::recursive_directory_iterator it( dir ), last;

	for( ; it != last; ++it )
	{
		std::string header = it->path().generic_string().substr( n+1 );

		std::ifstream is( it->path().c_str() );

		scan_header_dependencies( header, is, deps, from );
	}

	for( std::map< std::string, std::set< std::string > >::iterator i = deps.begin(); i != deps.end(); ++i )
	{
		if( i->first == module ) continue;

		std::cout << i->first << ":\n";

		for( std::set< std::string >::iterator j = i->second.begin(); j != i->second.end(); ++j )
		{
			std::cout << "    <" << *j << ">\n";

			std::set< std::string > const & f = from[ *j ];

			for( std::set< std::string >::const_iterator k = f.begin(); k != f.end(); ++k )
			{
				std::cout << "        from <" << *k << ">\n";
			}
		}

		std::cout << "\n";
	}
}

int main( int argc, char const* argv[] )
{
	if( argc < 2 )
	{
		std::cerr << "Usage: boostdep <module> [<module>...]\n";
		return -1;
	}

	try
	{
		build_header_map();
	}
	catch( fs::filesystem_error const & x )
	{
		std::cout << x.what() << std::endl;
	}

	for( int i = 1; i < argc; ++i )
	{
		std::string module = argv[ i ];

		std::cout << "Dependencies for " << module << ":\n\n";

		try
		{
			scan_module_dependencies( module );
		}
		catch( fs::filesystem_error const & x )
		{
			std::cout << x.what() << std::endl;
		}
	}
}
